! The subroutine COVMA(IUNIT, T2, COVMAT2, IERR) reads the
! covariance matrix from the covariance matrix data file 
! (logical unit number= IUNIT), propagates the covariance
! information to the target epoch T2 and gives the resulting
! matrix (COVMAT2) and an error code IERR as it is described
! for example in the CLUSTER DDID.
! Possible error codes are:
!      0 - No error
!      1 - Target epoch too early. If the target epoch is only 
!          a bit too early (<2sec) this error message needs not to
!          occur since some checks allow a small tolerance.
!      2 - Target epoch too late. If the target epoch is only 
!          a bit too late (<2sec) this error message needs not to
!          occur since some checks allow a small tolerance.   
!      4 - IUNIT is out of range, error during file rewind or
!          empty file (probably IUNIT of not existing file)
!      7 - File is inconsistent, e.g.
!          -Additional empty lines at the end of the file
!          -Overlap or gap between the validity time spans of
!           two consecutive records
!          -Data record does not have the correct format
! If the error code is not 0 the returned covariance matrix
! is undefined.
!
! A sample of two covariance matrix data file records:
!
!  3  R  2001-01-26T08:42:54Z  2001-01-21T06:48:20Z  2001-01-26T00:59:28Z
!107  386.283563137
!107 -23508.796   7831.817    -65.727 -0.0216890 -0.0217772  5.1843976
!  10.13166D-020.30640D-02-.27891D-02-.50794D-060.23258D-060.33014D-07
!  20.30640D-020.15653D-010.39614D-020.94519D-060.81966D-06-.24674D-06
!  3-.27891D-020.39614D-020.25249D-010.49340D-05-.24744D-06-.45987D-06
!  4-.50794D-060.94519D-060.49340D-050.98795D-090.25872D-10-.91822D-10
!  50.23258D-060.81966D-06-.24744D-060.25872D-100.28916D-09-.35936D-11
!  60.33014D-07-.24674D-06-.45987D-06-.91822D-10-.35936D-110.13128D-10
!  3  R  2001-01-30T09:03:23Z  2001-01-26T00:59:28Z  2001-01-28T10:10:24Z
!108  391.041292512
!108 -23601.093   7848.984   -602.626 -0.0769601  0.0169544  5.1712222
!  10.28264D-020.82365D-02-.61010D-02-.11127D-050.20669D-06-.25229D-07
!  20.82365D-020.25029D-01-.18491D-01-.33794D-050.50692D-06-.11760D-06
!  3-.61010D-02-.18491D-010.16490D-010.29027D-05-.67751D-060.10008D-06
!  4-.11127D-05-.33794D-050.29027D-050.55482D-09-.13535D-110.18279D-10
!  50.20669D-060.50692D-06-.67751D-06-.13535D-110.41707D-090.13353D-11
!  6-.25229D-07-.11760D-060.10008D-060.18279D-100.13353D-110.22010D-11
!  3  R  2001-02-02T07:58:46Z  2001-01-28T10:10:24Z  2001-02-02T04:26:47Z
!
! Before calling this subroutine an existing covmat file has  
! has to be opened with statements like
!      OPEN (UNIT=IUNIT, STATUS='OLD', IOSTAT=IOSTAT, 
!     $     ACCESS= 'SEQUENTIAL', FORM= 'FORMATTED')
!      IF (IOSTAT.NE.0) THEN
!         ...
!      ENDIF
! 
! Subroutines called are 
!    COVTRA, ELLIPS->PARDER, JD2000

      SUBROUTINE COVMA(IUNIT, T2, COVMAT2, IERR)
!Declarations
      IMPLICIT NONE

      !Arguments
      DOUBLE PRECISION
     $   T2,          !(in)  Time in Modified Julian Date (MJD2000)
     $   COVMAT2(6,6) !(out) Resulting covariance matrix of state
     $                !for time T2
      INTEGER
     $   IERR,        !(out) Error code
     $                !      0 - No error
     $                !      1 - Time too early
     $                !      2 - Time too late
     $                !      4 - IUNIT is out of range,
     $                !          error during file rewind
     $                !          or empty file 
     $                !          (probably IUNIT of not existing file)
     $                !      7 - File is inconsistent
     $   IUNIT        !(in)  Logical unit number of covariance
     $                !      matrix file

      !other variables
      CHARACTER
     $   ENDTIME*20,  !Record end time (CCSDS time code A format)
     $   GENTIME*20,  !Record generation time (CCSDS time code A format)
     $   SRTTIME*20,  !Record start time (CCSDS time code A format)
     $   PRED_OR_RECON*1 !If record is predicted or recon.
      DOUBLE PRECISION
     $   COVMAT1(6,6),!Covariance matrix read from record
     $   DSE,         !seconds 
     $   DUMMY_EM,    !not used but returned by ELLIPS
     $   DUMMY_RX,    !not used but returned by ELLIPS
     $   DUMMY_RY,    !not used but returned by ELLIPS
     $   F(6,6),      !F(I,K)= derivative Y(I) wrt X(K)
     $   LAST_MJD_ENDTIME,
     $   MJD_ENDTIME, !Record end time [MJD2000/UTC]
     $   MJD_GENTIME, !Record generation time [MJD2000/UTC]
     $   MJD_SRTTIME, !Record start time [MJD2000/UTC]
     $   T1,          !reference time state and covmat 
     $   TIME_DIFF,   !T2-T1 [sec]
     $   X(6),        !S/C state at T1
     $   Y(6)         !S/C state at T2  (Keppler orbit propagation)
      INTEGER
     $   COUNT,
     $   I,J,         !loop control
     $   INIT,        !0 needed by CALL ELLIPS
     $   IOSTAT,      !IOSTAT from latest READ or REWIND statement
     $   KODE,        !4 needed by CALL ELLIPS
     $   LAST_IERR,   !IERR returned after previous call of COVMA
     $   LAST_IUNIT,  !IUNIT provided to COVMA at previous call
     $   NREC,        !read from data file but not used
     $   NSAT,        !read from data file but not used
     $   YE, MO, DA, HO, MI, SE !year, month, day, hour, minute, second
      LOGICAL
     $   CONSISTENCY_ERROR,   !
     $   END_OF_FILE, !.TRUE. if end of file reached 
     $   FILE_CONSISTENCY_CHECK,  !.TRUE. if consistency of complete
     $                            !file shall be checked
     $   FOUND,           !.TRUE. if at T2 valid covmat record found
     $   LOOP,        !loop control
     $   REWIND_REQUIRED, !.TRUE. if it is required to start with
     $                !read from beginning of covmat data file
     $   SEARCH       !loop control

      !Parameters
      DOUBLE PRECISION
     $   TWOSEC       !2.01 seconds [day]
                      !Some times are only given with an accuracy
                      ! of 1 second in the data file. Therefore during
                      !several comparisons a margin of TWOSEC is used

      PARAMETER (
     $   TWOSEC=2.01D0/86400.0D0
     $   )

      !SAVE variables
      SAVE
     $   COUNT,
     $   COVMAT1,
     $   LAST_IERR,
     $   LAST_IUNIT,
     $   LAST_MJD_ENDTIME,
     $   MJD_ENDTIME,
     $   MJD_GENTIME,
     $   MJD_SRTTIME,
     $   NREC,
     $   NSAT,
     $   PRED_OR_RECON,
     $   X

      DATA
     $   LAST_IERR  /0/,
     $   LAST_IUNIT /0/

      !COMMON block CPOEMS
      !The subroutines ELLIPS and PARDER use some parameters
      !for earth satellite dynamics which are initialized in
      !this subroutine and transferred via the common block
      !CPOEMS. To do this via the common block is not nice but
      !has the advantage that the two orbit library subroutines
      !can be used without any modification.
      DOUBLE PRECISION XMU, WMU, WMUIN, XMUM, XMUS
      COMMON /CPOEMS/ XMU,WMU,WMUIN,XMUM,XMUS
!End declarations

      !Initialize common block CPOEMS
      !XMU = CENTRAL EARTH POTENTIAL, KM**3/SEC**2
      !XMUM, XMUS are not needed here
      XMU = 398600.440 D0
      WMU = DSQRT(XMU)
      WMUIN=1.D0/WMU
     
      IERR= 0

      IF( (IUNIT.NE.LAST_IUNIT) .OR. (LAST_IERR.EQ.7) ) THEN
         FILE_CONSISTENCY_CHECK=.TRUE.
      ELSE
         FILE_CONSISTENCY_CHECK=.FALSE.
      ENDIF   

      IF( (IUNIT.NE.LAST_IUNIT) .OR. (LAST_IERR.GT.2) ) THEN
         !New search from beginning of new file required
         COUNT= 0
         FOUND= .FALSE.
         REWIND_REQUIRED=.TRUE.
         SEARCH= .TRUE.
         LAST_IUNIT= IUNIT
      ELSE
         IF ((T2.GE.MJD_SRTTIME-TWOSEC) .AND. 
     $        (T2.LE.MJD_ENDTIME-TWOSEC)) THEN
            !T2 is covered by the record read during previous
            !call. If T2 is very close (TWOSEC) to MJD_ENDTIME
            !it is preferred to read and use already the next 
            !record. If there is no further record (END_OF_FILE)
            !it is decided later in the program to extend the
            !validity of the current record
            !until MJD_ENDTIME+TWOSEC                    (**1**)
            FOUND= .TRUE.
            REWIND_REQUIRED= .FALSE.
            SEARCH= .FALSE.
         ELSE IF (T2.LT.MJD_SRTTIME-TWOSEC) THEN
            !New search from beginning of file required 
            !(requires rewind)
            !Remark: There is no backwards search of records 
            !implemented
            COUNT= 0
            FOUND= .FALSE.
            REWIND_REQUIRED= .TRUE.
            SEARCH= .TRUE.
         ELSE
            !T2 GT. MJD_ENDTIME-TWOSEC
            !Forward search, read next records in
            FOUND= .FALSE.
            REWIND_REQUIRED= .FALSE.
            SEARCH= .TRUE.
         ENDIF               
      ENDIF
      

      !check complete file if it is consistent.
      !Therefore every record is read once 
      IF (FILE_CONSISTENCY_CHECK) THEN
         REWIND(UNIT=IUNIT,IOSTAT=IOSTAT)
         IF (IOSTAT.NE.0) THEN 
            IERR= 4
            LOOP= .FALSE.
            SEARCH= .FALSE.
            CONSISTENCY_ERROR= .FALSE.
         ELSE
            LOOP=.TRUE.
         ENDIF

         DO WHILE (LOOP)
            READ (UNIT=IUNIT, FMT='(I3,2X,A1,2X,2(A20,2X),A20)',
     $           IOSTAT=IOSTAT, END=10) NSAT, PRED_OR_RECON,
     $           GENTIME, SRTTIME, ENDTIME
            END_OF_FILE= .FALSE.
            GOTO 11
 10         END_OF_FILE= .TRUE.
 11         CONTINUE

            IF (END_OF_FILE) THEN
               !End of file reached without detecting any error
               LOOP= .FALSE.
               CONSISTENCY_ERROR= .FALSE.
               COUNT= 0
            ELSE
               READ(SRTTIME, IOSTAT=IOSTAT, FMT='(I4,1X,I2.2,1X,'//
     $              'I2.2,1X,I2.2,1X,I2.2,1X,I2.2,1X)') 
     $              YE, MO, DA, HO, MI, SE
               IF (IOSTAT.EQ.0) THEN
                  DSE=DBLE(SE)
                  CALL JD2000(MJD_SRTTIME, YE, MO, DA, HO, MI, DSE)
                  READ(ENDTIME, IOSTAT=IOSTAT,
     $                 FMT='(I4,1X,I2.2,1X,I2.2,1X,I2.2,1X,I2.2,'//
     $                 '1X,I2.2,1X)')
     $                 YE, MO, DA, HO, MI, SE
               ENDIF
               IF (IOSTAT.EQ.0) THEN
                  DSE=DBLE(SE)
                  CALL JD2000(MJD_ENDTIME, YE, MO, DA, HO, MI, DSE)
                  READ(GENTIME, IOSTAT=IOSTAT, 
     $                 FMT='(I4,1X,I2.2,1X,I2.2,1X,I2.2,1X,I2.2,'//
     $                 '1X,I2.2,1X)')
     $                 YE, MO, DA, HO, MI, SE
               ENDIF
               IF (IOSTAT.EQ.0) THEN
                  DSE=DBLE(SE)
                  CALL JD2000(MJD_GENTIME, YE, MO, DA, HO, MI, DSE) 
                  READ (UNIT=IUNIT, FMT='(I3,F15.9)', IOSTAT=IOSTAT)
     $                 NREC, T1
               ENDIF
               IF (IOSTAT.EQ.0) THEN
                  READ (UNIT=IUNIT, FMT='(I3,3F11.3,3F11.7)', 
     $                 IOSTAT=IOSTAT)
     $                 NREC,(X(I),I=1,6)
               ENDIF
               IF (IOSTAT.EQ.0) THEN
                  DO J=1,6
                     READ (UNIT=IUNIT, FMT='(I3,6(D11.5))',
     $                    IOSTAT=IOSTAT)
     $                    NREC,(COVMAT1(I,J),I=1,6)
                     IF (IOSTAT.NE.0) THEN                     
                        GOTO 15 !EXIT loop
                     ENDIF
                  ENDDO
 15               CONTINUE
               ENDIF

               IF (IOSTAT.EQ.0) THEN
                  !The record was read successfully
                  COUNT= COUNT + 1
                  IF (COUNT.EQ.1) THEN
                     LAST_MJD_ENDTIME= MJD_ENDTIME
                  ELSE IF (COUNT.GE.2) THEN
                     !Do some consistency checks
                     IF ( (MJD_SRTTIME + TWOSEC) .LT.
     $                    LAST_MJD_ENDTIME) THEN
                        !record validity start time earlier than
                        !previous end time;
                        !wrong sequence of records or overlap
                        IERR=7
                        LOOP=.FALSE.
                        CONSISTENCY_ERROR=.TRUE.
                     ELSE IF ( (MJD_SRTTIME - TWOSEC) .GT.
     $                       LAST_MJD_ENDTIME) THEN
                        !gap between records
                        IERR=7
                        LOOP=.FALSE.
                        CONSISTENCY_ERROR=.TRUE.
                     ENDIF
                     LAST_MJD_ENDTIME= MJD_ENDTIME
                  ENDIF
               ELSE
                  !I/O error during read, file inconsistent
                  IERR= 7
                  LOOP=.FALSE.
                  CONSISTENCY_ERROR=.TRUE.
               ENDIF
            ENDIF
         ENDDO
      ELSE
         CONSISTENCY_ERROR= .FALSE.
      ENDIF  !End file consistency check

      IF (CONSISTENCY_ERROR) THEN
         SEARCH= .FALSE.
      ENDIF

      IF (REWIND_REQUIRED) THEN
         REWIND(UNIT=IUNIT,IOSTAT=IOSTAT)
         IF (IOSTAT.NE.0) THEN 
            IERR= 4
            SEARCH= .FALSE.
         ENDIF
      ENDIF

      DO WHILE (SEARCH)

         READ (UNIT=IUNIT, FMT='(I3,2X,A1,2X,2(A20,2X),A20)',
     $        IOSTAT=IOSTAT, END=20) NSAT, PRED_OR_RECON,
     $        GENTIME, SRTTIME, ENDTIME
         END_OF_FILE= .FALSE.
         GOTO 21
 20      END_OF_FILE= .TRUE.
 21      CONTINUE

         IF (END_OF_FILE) THEN
            SEARCH=.FALSE.
            IF (COUNT.EQ.0) THEN
               !No data in file
               !Probably unit not connected to a proper file
               IERR=4    
            ELSE 
               !At least one valid record was read earlier
               IF (T2.GT.MJD_ENDTIME+TWOSEC) THEN
                  IERR=2
               ELSE
                  FOUND=.TRUE.
                  !T2 is covered by record read during previous call
                  !Normally this is already detected earlier in 
                  !the program if T2.GT.MJD_ENDTIME-TWOSEC.
                  !Please read comment marked with (**1**) above
               ENDIF
            ENDIF
         ELSE
            READ(SRTTIME, IOSTAT=IOSTAT, 
     $           FMT='(I4,1X,I2.2,1X,I2.2,1X,I2.2,1X,I2.2,1X,I2.2,1X)') 
     $           YE, MO, DA, HO, MI, SE
            IF (IOSTAT.EQ.0) THEN
               DSE=DBLE(SE)
               CALL JD2000(MJD_SRTTIME, YE, MO, DA, HO, MI, DSE)
               READ(ENDTIME, IOSTAT=IOSTAT,
     $            FMT='(I4,1X,I2.2,1X,I2.2,1X,I2.2,1X,I2.2,1X,I2.2,1X)')
     $            YE, MO, DA, HO, MI, SE
            ENDIF
            IF (IOSTAT.EQ.0) THEN
               DSE=DBLE(SE)
               CALL JD2000(MJD_ENDTIME, YE, MO, DA, HO, MI, DSE)
               READ(GENTIME, IOSTAT=IOSTAT, 
     $            FMT='(I4,1X,I2.2,1X,I2.2,1X,I2.2,1X,I2.2,1X,I2.2,1X)')
     $            YE, MO, DA, HO, MI, SE
            ENDIF
            IF (IOSTAT.EQ.0) THEN
               DSE=DBLE(SE)
               CALL JD2000(MJD_GENTIME, YE, MO, DA, HO, MI, DSE) 
               READ (UNIT=IUNIT, FMT='(I3,F15.9)', IOSTAT=IOSTAT)
     $           NREC, T1
            ENDIF
            IF (IOSTAT.EQ.0) THEN
               READ (UNIT=IUNIT, FMT='(I3,3F11.3,3F11.7)', 
     $              IOSTAT=IOSTAT)
     $              NREC,(X(I),I=1,6)
            ENDIF
            IF (IOSTAT.EQ.0) THEN
               DO J=1,6
                  READ (UNIT=IUNIT, FMT='(I3,6(D11.5))', IOSTAT=IOSTAT)
     $                 NREC,(COVMAT1(I,J),I=1,6)
                  IF (IOSTAT.NE.0) THEN                     
                     GOTO 30 !EXIT loop
                  ENDIF
               ENDDO
 30            CONTINUE
            ENDIF

            IF (IOSTAT.EQ.0) THEN
               !The record was read successfully
               COUNT= COUNT + 1

               IF (COUNT.EQ.1) THEN
                  IF (T2.LT.MJD_SRTTIME-TWOSEC) THEN
                     IERR= 1
                     SEARCH= .FALSE.
                     FOUND= .FALSE.
                  ENDIF
                  LAST_MJD_ENDTIME= MJD_ENDTIME
               ELSE IF (COUNT.GE.2) THEN
                  !Do some consistency checks
                  IF ( (MJD_SRTTIME + TWOSEC) .LT.
     $                 LAST_MJD_ENDTIME) THEN
                     !record validity start time earlier than previous
                     !end time; wrong sequence of records or overlap
                     IERR=7
                     SEARCH= .FALSE.
                     FOUND= .FALSE.
                  ELSE IF ( (MJD_SRTTIME - TWOSEC) .GT.
     $                    LAST_MJD_ENDTIME) THEN
                     !gap between records
                     IERR=7
                     SEARCH= .FALSE.
                     FOUND= .FALSE.
                  ENDIF
                  LAST_MJD_ENDTIME= MJD_ENDTIME
               ENDIF

               !In the if once 2*TWOSEC is used. See also comment
               !marked with (**1**) above. This is a consequence
               !of the (TWOSEC) earlier use of the next record.
               !Therefore a further margin of TWOSEC has to be
               !used here.
               IF ((T2.GE.MJD_SRTTIME-2.0D0*TWOSEC) .AND. 
     $              (T2.LE.MJD_ENDTIME+TWOSEC)) THEN
                  FOUND= .TRUE.
                  SEARCH= .FALSE.
               ENDIF               
            ELSE
               IERR= 7  !I/O error during read, file inconsistent
               SEARCH= .FALSE.
            ENDIF                     
         ENDIF

      ENDDO

      !Now do covariance propagation
      IF (FOUND) THEN
         INIT= 0
         KODE= 4
         TIME_DIFF= (T2 - T1)*86400.0D0
         CALL ELLIPS (X, INIT, KODE, TIME_DIFF, Y, F, DUMMY_RX,
     $        DUMMY_RY, DUMMY_EM)
         !               t
         ! C   =  F  C  F
         !  t2        t1
         CALL COVTRA(COVMAT1, F, COVMAT2)
      ENDIF

      LAST_IERR= IERR
      IF (IERR.NE.0) THEN
         COVMAT2(1,1)=-1.0D-99
         COVMAT2(2,2)=-1.0D-99
         COVMAT2(3,3)=-1.0D-99
         COVMAT2(4,4)=-1.0D-99
         COVMAT2(5,5)=-1.0D-99
         COVMAT2(6,6)=-1.0D-99
      ENDIF

      END
